//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de
//**************************************************************************************************

//! \file  manip.cpp
//! \brief NManip: manipulate data (select, average, rearrange, ...)

#include "defs.hpp"

#include <boost/format.hpp>

#include "../trivia/vector_ops.hpp"
#include "../readplus/ask.hpp"

#include "olf.hpp"
#include "mem.hpp"
#include "fsel.hpp"
#include "loop.hpp"
#include "jsel.hpp"
#include "slice.hpp"
#include "obj.hpp"
#include "expr.hpp" // for sort, symmetrize
#include "manip.hpp"
#include "xax_lex.hpp"

using boost::format;

// local routines:
namespace NManip {
    void ScaRemoveConstantZ( COlo *f );
}


//**************************************************************************************************
//*  restrict forthcoming operations
//**************************************************************************************************

//! Makes some spectra unmutable.

void NManip::freeze_slices()
{
    FileIterator fiter(SFSel::instance()->sel());
    string sel;
    sel = wask( "Protect which spectra from forthcoming operations" );
    if ( sel=="" )
        return;
    RNode Tsel = user_xaxparse( sel.c_str() );
    while ( COlo *fio = fiter.next() ) {
        RObjVecInt vsel = Tsel->to_index_list( fio->nJ() );
        if (!vsel)
            throw S("not a valid index list");
        for ( int j=0; j<fio->nJ(); ++j )
            fio->V[j]->frozen = triv::contains( vsel->v, j );
        fio->log_action( "m/ " + sel );
    }
}

//! Makes all spectra mutable.

void NManip::unfreeze_all_slices()
{
    FileIterator fiter(SFSel::instance()->sel());
    while ( COlo *fio = fiter.next() ) {
        for ( int j=0; j<fio->nJ(); ++j )
            fio->V[j]->frozen = false;
        fio->log_action( "m/-" );
    }
}


//**************************************************************************************************
//*  manipulations on data points
//**************************************************************************************************

//! Delete (or retain) points according to list of indices.

void NManip::points_select( bool sel_del )
{
    FileIterator fiter(SFSel::instance()->selD());
    static string sel = "";
    sel = wask( string( sel_del ? "Delete" : "Retain" ) + " which points", sel );
    if ( sel=="" )
        return;
    RNode Tsel = user_xaxparse( sel.c_str() );

    while ( const COld *fin = fiter.nextD() ) {
        POld fout( fin->new_POld() );
        fout->log_action( "mp" + string( sel_del ? "d" : "r" ) + " " + sel );
        for ( int j=0; j<fin->nJ(); j++ ) {
            const CSpec *sin = fin->VS(j);
            PSpec sout( new CSpec );
            sout->z = sin->z;
            RObjVecInt vsel = Tsel->to_index_list( sin->size() );
            if (!vsel)
                throw S("not a valid index list");
            for ( int i=0; i<sin->size(); i++ ) {
                if ( sel_del ^ triv::contains(vsel->v, i) ) {
                    if ( sin->has_dy() )
                        sout->push_xyd( sin->x[i], sin->y[i], sin->dy[i] );
                    else
                        sout->push_xy( sin->x[i], sin->y[i] );
                }
            }
            if ( sout->size() )
                fout->V.push_back( move(sout) );
        }
        if ( !fout->V.size() )
            throw S("no data point left");
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Bin points.

void NManip::points_rebin()
{
    FileIterator fiter(SFSel::instance()->selD());

    static string sel;
    sel = wask( "Start bins at points", sel );
    if ( sel=="" )
        return;
    RNode Tsel = user_xaxparse( sel.c_str() );

    while ( const COld *fin = fiter.nextD() ) {
        POld fout( fin->new_POld() );
        fout->log_action("mpa " + sel);

        for ( int j=0; j<fin->nJ(); j++ ) {
            const CSpec *sin = fin->VS(j);
            RObjVecInt vsel = Tsel->to_index_list( sin->size() );
            if (!vsel)
                throw S("not a valid index list");
            if ( !vsel->size() )
                throw S("BUG: unexpected empty bin list");
            vector<int> breaks = vsel->v;
            if ( breaks[0]!=0 )
                throw S("First bin must start at 0");
            breaks.push_back( sin->size() );

            PSpec sout = sin->binned( breaks );

            fout->V.push_back( move(sout) );
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Bin points.

void NManip::points_rebin_by_factor()
{
    FileIterator fiter(SFSel::instance()->selD());
    static int ng=2;
    ng = iask( "Number of points per bin", ng );
    if ( ng<1 )
        throw S("invalid choice");

    while ( const COld *fin = fiter.nextD() ) {
        POld fout( fin->new_POld() );
        fout->log_action( "mpaf " + S(ng) );

        for ( int j=0; j<fin->nJ(); j++ ) {
            const CSpec* sin = fin->VS(j);

            int nout = sin->size() / ng;
            if ( nout<1 )
                throw "not enough points in spectrum " + S(j);
            vector<int> breaks( nout+1 );
            for ( int iout=0; iout<nout+1; ++iout )
                breaks[iout] = iout*ng;

            PSpec sout = sin->binned( breaks );

            fout->V.push_back( move(sout) );
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Bin points to keep errors below a certain bound.

void NManip::points_rebin_by_err( const string& subcmd )
{
    FileIterator fiter(SFSel::instance()->selD());
    bool mode_abs;
    if      ( subcmd=="a" )
        mode_abs = true;
    else if ( subcmd=="r" )
        mode_abs = false;
    else
        throw S("invalid subcommand");

    static double errbound;
    errbound = dask( "Keep error below", errbound );

    while ( const COld *fin = fiter.nextD() ) {
        POld fout( fin->new_POld() );
        string groupinfo;
        for ( int j=0; j<fin->nJ(); j++ ) {
            const CSpec *sin = fin->VS(j);
            PSpec sout( new CSpec() );
            sout->z = sin->z;
            if ( !sin->has_dy() )
                throw S("no errors available");
            double xg = 0, yg = 0, vg = 0;
            int ng = 0, istart=0;
            for ( int i=0; i<sin->size(); ++i ) {
                if ( !mode_abs && sin->y[i]<0 )
                    throw ("negative counts not compatible with relative error bounds");
                xg += sin->x[i];
                yg += sin->y[i];
                vg += SQR( sin->dy[i] );
                ++ng;
                double dg = sqrt(vg);
                if ( (  mode_abs && dg/ng<errbound ) ||
                     ( !mode_abs && yg>0 && dg/yg<errbound ) )
                {
                    sout->push_xyd( xg/ng, yg/ng, dg/ng );
                    groupinfo += S(istart) + ",";
                    xg = 0; yg = 0; vg = 0; ng = 0;
                    istart = i+1;
                }
            }
            if ( sout->size() )
                fout->V.push_back( move(sout) );
        }
        string docline = "mpae" + subcmd + " " + S(errbound);
        if ( fin->nJ()==1 )
            docline += " #=> groups " + groupinfo.substr(0,groupinfo.size()-1);
        fout->log_action( docline );
        if ( fout->nJ() )
            SMem::instance()->mem_store( move(fout), fiter.k() );

    }
}


//! Break into spectra, by fixed factor.

void NManip::points_break_by_factor()
{
    FileIterator fiter(SFSel::instance()->selD());
    static int ng;
    ng = iask( "Number of points per spectrum", ng );
    if ( ng<1 )
        throw S("invalid choice");

    while ( const COld *fin = fiter.nextD() ) {
        POld fout( fin->new_POld() );
        fout->log_action( "mpbf " + S(ng) );
        fout->ZCo.push_back( CCoord("no-in-break", ""));

        for ( int j=0; j<fin->nJ(); j++ ) {
            const CSpec* sin = fin->VS(j);
            int jsub = 0;
            for( int i=0; i<sin->size(); i+=ng ){
                PSpec sout( new CSpec() );
                sout->z = sin->z;
                sout->z.push_back( RObjInt( new CObjInt(jsub++) ) );
                for( int ii=i; ii<sin->size() && ii<i+ng; ++ii ){
                    if( sin->has_dy() )
                        sout->push_xyd( sin->x[ii], sin->y[ii], sin->dy[ii] );
                    else
                        sout->push_xy( sin->x[ii], sin->y[ii] );
                }
                fout->V.push_back( move(sout) );
            }
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Sort spectrum according to given expression.

void NManip::points_sort()
{
    FileIterator fiter(SFSel::instance()->selD());
    string expr = sask("Sort points according to");
    if (expr=="") return;
    RNode T = user_xaxparse( expr.c_str() );
    if ( T->has_dummy() )
        throw S("dummy argument t not allowed when operating on data");

    while ( const COld *fin = fiter.nextD() ) {
        int k = fiter.k();
        POld fout( fin->new_POld() );
        fout->log_action("mpo "+expr);

        for ( int j=0; j<fin->nJ(); j++ ) {
            const CSpec *sin = fin->VS(j);
            PSpec sout( new CSpec );
            sout->z = sin->z;

            int n = sin->size();
            vector<double> v(n);
            T->tree_vec_val( &v, 0, k, j );
            vector<size_t> P = triv::sorted_indices( v );
            sout->resize( n, sin->has_dy() );
            for (int i=0; i<n; ++i) {
                sout->x[i] = sin->x[P[i]];
                sout->y[i]  = sin->y[P[i]];
                if ( sin->has_dy() )
                    sout->dy[i] = sin->dy[P[i]];
            }
            fout->V.push_back( move(sout) );
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Average points that have the same x.

void NManip::points_rebin_duplicates()
{
    FileIterator fiter(SFSel::instance()->selD());
    COld fout;
    while ( const COld *fin = fiter.nextD() ) {
        POld fout( fin->new_POld() );
        fout->log_action("mpq # average points when x equal");

        for ( int j=0; j<fin->nJ(); j++ ) {
            const CSpec *sin = fin->VS(j);
            PSpec sout( new CSpec );
            sout->z = sin->z;

            int n = sin->size();
            int ii=0;
            for ( int i=0; i<n; i=ii ) {
                double xg = sin->x[i];
                double yg = sin->y[i];
                double eg = NAN;
                if ( sin->has_dy() )
                    eg = SQR(sin->dy[i]);
                int ng = 1;
                for ( ii=i+1; ii<n; ++ii ) {
                    if (sin->x[ii] != xg) break;
                    yg += sin->y[ii];
                    if ( sin->has_dy() )
                        eg += SQR( sin->y[ii] );
                    ++ng;
                }
                if ( sin->has_dy() )
                    sout->push_xyd(xg, yg/ng, sqrt(eg)/ng );
                else
                    sout->push_xy(xg, yg/ng);
            }
            fout->V.push_back( move(sout) );
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Symmetrize spectrum.

void NManip::points_symmetrize()
{
    FileIterator fiter(SFSel::instance()->selD());
    static string expr = "0";
    expr = sask( "Assume symmetry with mirror positioned at", expr );
    if ( expr=="" ) return;
    RNode T = user_xaxparse( expr.c_str() );

    while ( const COld *fin = fiter.nextD() ) {
        int k = fiter.k();
        POld fout( fin->new_POld() );
        fout->log_action("mpsym "+expr);

        for (int j=0; j<fin->nJ(); j++) {
            int n, i, il, ih, nl, nh;
            const CSpec *sin = fin->VS(j);
            PSpec sout( new CSpec );
            sout->z = sin->z;
            if ( !triv::is_ascending( sin->x ) )
                throw S("not ascending");
            double step;
            if ( !triv::is_equidist( &step, sin->x ) )
                throw S("not equidistant");
            double centre = T->tree_point_dbl( k, j );
            il = (int) ( (centre - sin->x[0])/step );
            ih = il + 1;
            n = sin->size();
            nl = il + 1;
            nh = n - ih;
            if ( nl<1 || nh<1 )
                throw S("nothing to symmetrize");
            for (i=0; i<(nl<nh?nl:nh); ++i) {
                if ( sin->has_dy() )
                    sout->push_xyd( ( -sin->x[il-i] + sin->x[ih+i] ) / 2,
                                    ( sin->y[il-i] + sin->y[ih+i] ) / 2,
                                    sqrt( SQR(sin->dy[il-i]) + SQR(sin->dy[ih+i]) ) / 2 );
                else
                    sout->push_xy( ( -sin->x[il-i] + sin->x[ih+i] ) / 2,
                                   ( sin->y[il-i] + sin->y[ih+i] ) / 2 );
            }
            if ( nl<nh ) {
                for (i=nl; i<nh; ++i)
                    if ( sin->has_dy() )
                        sout->push_xyd( sin->x[ih+i], sin->y[ih+i], sin->dy[ih+i] );
                    else
                        sout->push_xy( sin->x[ih+i], sin->y[ih+i] );
            } else {
                for (i=nh; i<nl; ++i)
                    if ( sin->has_dy() )
                        sout->push_xyd( -sin->x[il-i], sin->y[il-i], sin->dy[il-i] );
                    else
                        sout->push_xy( -sin->x[il-i], sin->y[il-i] );
            }
            sout->z = sin->z;
            fout->V.push_back( move(sout) );
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}

//! Remove error bars.

void NManip::points_remove_err()
{
    FileIterator fiter(SFSel::instance()->selD());
    while ( const COld *fin = fiter.nextD() ) {
        POld fout( fin->new_POld() );
        fout->log_action("mpe-");
        for (int j=0; j<fin->nJ(); j++) {
            const CSpec *sin = fin->VS(j);
            PSpec sout( new CSpec );
            sout->z = sin->z;
            sout->x = sin->x;
            sout->y = sin->y;
            fout->V.push_back( move(sout) );
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//**************************************************************************************************
//*  manipulations on spectra
//**************************************************************************************************

//! Remove z variable that does not genusy with j.

void NManip::ScaRemoveConstantZ( COlo *fio )
{
    for ( int iz=fio->nZ()-1; iz>=0; --iz ) {
        int nj = fio->nJ();
        if ( nj<1 )
            throw S("file contains no spectrum");
        double z = fio->z(0,iz)->to_r();
        bool z_is_const = true;
        for ( int j=1; j<nj; ++j ) {
            if ( fio->z(j,iz)->to_r() != z ) {
                z_is_const = false;
            }
        }
        if ( z_is_const ) {
            printf( " constant z%i becomes r%i\n", iz, (int)fio->RPar.size() );
            fio->RPar.push_back( CParam( fio->ZCo[iz], z ) );
            fio->remove_z( iz );
        }
    }
}


//! Delete (or retain) spectra according to list of indices.

void NManip::slices_select( bool sel_del )
{
    vector<int> JSel, JSelSorted;
    static string jSel = "";
    JSelAsk( string( sel_del ? "Delete" : "Retain" ) + " which spectra", JSel );

    FileIterator fiter(SFSel::instance()->sel());
    while ( const COlo *fin = fiter.next() ) {
        POlo fout( fin->new_POlo() );
        // JSel.evaluate( 0, fin->nJ()-1 );
        fout->log_action( "ms" + string( sel_del ? "d" : "r" ) +
                              " " + triv::indices_to_s(JSel) );
        for ( int j=0; j<fin->nJ(); ++j ) {
            if ( sel_del ^ triv::contains(JSel, j ) ) {
                fout->V.push_back( fin->copy_slice( j ) );
            }
        }
        ScaRemoveConstantZ( fout.get() );
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Bin spectra.

void NManip::slices_rebin()
{
    FileIterator fiter(SFSel::instance()->sel());
    vector<int> JSel;
    static string jSel = "";
    JSelAsk( "Start groups at spectra", JSel );
    while ( const COlo *fin = fiter.next() ) {
        if( fin->nJ()<=1 )
            throw S("Only one slice in file");
        const COld *fd = dynamic_cast<const COld*>( fin );
        const COlc *fc = dynamic_cast<const COlc*>( fin );
        POlo fout( fin->new_POlo() );
        // JSel.evaluate( 0, fin->nJ()-1 );
        if ( JSel.size()<1 || JSel[0]!=0 )
            throw S("spectrum selection must contain 0");
        vector<int> BreakPoints;
        copy_if( JSel.begin(), JSel.end(), back_inserter(BreakPoints),
                 [&](int j){ return j<fin->nJ(); } );
        BreakPoints.push_back( fin->nJ() );
        fout->log_action("msa " + triv::indices_to_s(JSel));
        for ( int iv=0; iv<BreakPoints.size()-1; iv++ ) {
            int ji = BreakPoints[iv];
            int jf = BreakPoints[iv+1];
            int mj = jf-ji;
            if( mj<=0 )
                throw S("Empty group");
            int nz = fin->nZ();
            vector<double> zout( nz );
            for ( int iz=0; iz<nz; ++iz ) {
                double z = 0;
                for ( int jj=ji; jj<jf; jj++ ) {
                    auto pz = PCAST<const CObjNum>(fin->z(jj,iz));
                    if( !pz )
                        throw "z"+S(iz)+" is not a number";
                    z += pz->to_r();
                }
                zout[iz] = z/mj;
            }
            if ( fd ) {
                PSpec sout( new CSpec() );
                bool in_with_dy = fd->VS(ji)->has_dy();
                bool out_with_dy = in_with_dy || mj>1;
                int n = fd->nPts(ji);
                sout->resize( n, out_with_dy );
                // x grids must be equal:
                sout->x = fd->VS(ji)->x;
                for ( int jj=ji+1; jj<jf; jj++ ) {
                    if ( fd->VS(jj)->has_dy() != in_with_dy )
                        throw S("some slices have dy, others not");
                    if ( fd->nPts(jj) != n )
                        throw "grids differ in size: ni[,"+S(jj)+"]="+S(fd->nPts(jj))+
                            " vs ni[,"+S(ji)+"]="+S(n);
                    for ( int i=0; i<fd->nPts(jj); ++i ) {
                        if ( fd->VS(jj)->x[i] != sout->x[i] )
                            throw "grids differ: x[,"+S(jj)+","+S(i)+"]="+S(fd->VS(jj)->x[i])+
                                " vs x[,"+S(ji)+","+S(i)+"]="+S(sout->x[i]);
                    }
                }
                // bin y:
                for ( int i=0; i<n; i++ ) {
                    double ym = 0;
                    for ( int jj=ji; jj<jf; jj++)
                        ym += fd->VS(jj)->y[i];
                    if ( mj!=jf-ji || mj<1 )
                        throw S("BUG: inconsistent number of binned spectra");
                    ym /= mj;
                    sout->y[i] = ym;
                    if ( out_with_dy ) {
                        // TODO needs justification / documentation
                        if (mj<2) {
                            if ( !in_with_dy )
                                throw S("BUG: should not compute variance for case mj=1");
                            sout->dy[i] = fd->VS(ji)->dy[i];
                        } else {
                            double vm_src = 0, vm_grp = 0;
                            for ( int jj=ji; jj<jf; jj++) {
                                if( fd->VS(jj)->has_dy() )
                                    vm_src += SQR(fd->VS(jj)->dy[i]);
                                vm_grp += SQR(fd->VS(jj)->y[i]-ym);
                            }
                            double vm = vm_src/mj/mj;
                            if( mj>1 )
                                vm += vm_grp / (mj-1);
                            sout->dy[i] = sqrt(vm);
                        }
                    }
                }
                for ( int iz=0; iz<nz; ++iz )
                    sout->z.push_back( PObjDbl( new CObjDbl(zout[iz]) ) );
                fout->V.push_back( move(sout) );
            } else if ( fc ) {
                PCurve cout( new CCurve );
                int np = fc->nP;
                vector<double> p(np);
                for ( int ip=0; ip<np; ip++ ) {
                    int mj = 0;
                    double pm = 0;
                    for ( int jj=ji; jj<jf; jj++) {
                        pm += fc->VC(jj)->P[ip];
                        ++mj;
                    }
                    p[ip] = mj==0 ? 0 : pm/mj;
                }
                cout->P = p;
                for ( int iz=0; iz<nz; ++iz )
                    cout->z.push_back( PObjDbl( new CObjDbl(zout[iz]) ) );
                cout->ParAttr.resize( np, 'u' );
                fout->V.push_back( move(cout) );
            }
        }
        ScaRemoveConstantZ( fout.get() );
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Merge spectra, concatenating points.

void NManip::slices_merge()
{
    FileIterator fiter(SFSel::instance()->selD());
    vector<int> JSel;
    static string jSel = "";
    JSelAsk( "Start groups at spectra", JSel );

    while ( const COld *fin = fiter.nextD() ) {
        POld fout( fin->new_POld() );
        // JSel.evaluate( 0, fin->nJ()-1 );
        if (JSel.size()<1 || JSel[0]!=0)
            throw S("spectrum selection must contain 0");
        fout->log_action("msj " + triv::indices_to_s(JSel));
        cout << "WARNING: z just taken from first spectrum of each group\n";
        for ( int iv=0; iv<JSel.size(); iv++ ) {
            int ji, jf;
            ji = JSel[iv];
            jf = ( iv<JSel.size()-1 ? JSel[iv+1] : fin->nJ() );
            PSpec sout( new CSpec( *(fin->VS(ji)) ) );
            bool with_dy = fin->VS(ji)->has_dy();
            for ( int jj=ji+1; jj<jf; jj++) {
                if ( fin->VS(jj)->has_dy() != with_dy )
                    throw S("some spectra have error bars, others not");
                for ( int i=0; i<fin->VS(jj)->size(); ++i )
                    if ( with_dy )
                        sout->push_xyd( fin->VS(jj)->x[i], fin->VS(jj)->y[i], fin->VS(jj)->dy[i]);
                    else
                        sout->push_xy( fin->VS(jj)->x[i], fin->VS(jj)->y[i] );
            }
            fout->V.push_back( move(sout) );
        }
        ScaRemoveConstantZ( fout.get() );
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! After each span, insert given number of copies.

void NManip::slices_spawn()
{
    FileIterator fiter(SFSel::instance()->sel());
    static int njj = 2;
    int njjIn = iask("output will have how many copies of input spectrum", njj);
    if (njjIn<1) return;
    njj = njjIn;

    while ( const COlo *fin = fiter.next() ) {
        POlo fout( fin->new_POlo() );
        fout->log_action( "ms* " + S(njj) );

        fout->ZCo.insert( fout->ZCo.begin(), CCoord("no-in-spawn", ""));

        for ( int jj=0; jj<njj; ++jj ) {
            for ( int j=0; j<fin->nJ(); j++ ) {
                PSlice sout = fin->copy_slice( j );
                sout->z.insert( sout->z.begin(), PObjInt( new CObjInt( jj ) ) );
                fout->V.push_back( move(sout) );
            }
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Exchange x and z.

void NManip::exchange_x_z()
{
    FileIterator fiter(SFSel::instance()->selD());
    while ( const COld *fin = fiter.nextD() ) {
        int izco = fin->ZCo.size() - 1;
        if ( izco==(int)-1 )
            throw S("no input z coordinate");
        POld fout( fin->new_POld() );
        fout->log_action(
            "msx # exchanging x=" + fin->xco.str_compact() +
            " with z" + S(izco) + "=" + fin->ZCo[izco].str_compact() );
        fout->xco = fin->ZCo[izco];
        fout->ZCo[izco] = fin->xco;

        if ( fin->nJ()<=0 )
            throw S("no spectra in file");

        vector<double> zcommon, xcommon;
        int ji, jf;

        for ( ji=0; ji<fin->nJ(); ) {
            zcommon.clear();
            for ( int i=0; i<izco; ++i)
                zcommon.push_back(fin->V[ji]->z[i]->to_r());

            for ( jf=ji+1; jf<fin->nJ(); jf++ )
                for ( int i=0; i<izco; ++i)
                    if (fin->V[jf]->z[i]->to_r() != zcommon[i])
                        goto end_group;
        end_group:

            // build common grid:
            xcommon.clear();
            for ( int j=ji; j<jf; ++j )
                for ( int i=0; i<fin->VS(j)->size(); ++i )
                    xcommon.push_back( fin->VS(j)->x[i] );
            sort( xcommon.begin(), xcommon.end() );
            triv::make_unique(&xcommon, 1e-180, 1e-20);

            printf("group of spectra %i .. %i has %i different x\n", ji, jf-1, (int)xcommon.size());

            int ii = 0; // a guess
            for ( int i=0; i<xcommon.size(); ++i ) {
                PSpec sout( new CSpec );

                for( double z: zcommon )
                    sout->z.push_back( PObjDbl( new CObjDbl(z) ) );
                sout->z.push_back( PObjDbl( new CObjDbl(xcommon[i]) ) );

                bool with_dy = true;
                for ( int j=ji; j<jf; ++j ) {
                    if ( !( fin->VS(j)->has_dy() ) ) {
                        with_dy = false;
                        break;
                    }
                }
                for ( int j=ji; j<jf; ++j ) {
                    const CSpec* sin = fin->VS( j );
                    if ( sin->x[ii]!=xcommon[i] ) { // guess failed
                        for (ii=0; ii<sin->size(); ++ii)
                            if ( sin->x[ii]==xcommon[i] )
                                break;
                    }
                    if ( ii<sin->size() ) {
                        if ( with_dy )
                            sout->push_xyd( sin->z[izco]->to_r(), sin->y[ii], sin->dy[ii] );
                        else
                            sout->push_xy( sin->z[izco]->to_r(), sin->y[ii] );
                    }
                }
                fout->V.push_back( move(sout) );
                ++ii;
            }
            ji = jf;
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Sort slices according to expression.

void NManip::slices_sort()
{
    FileIterator fiter(SFSel::instance()->sel());
    string expr = sask("Sort spectra according to");
    if (expr=="") return;
    RNode T = user_xaxparse( expr.c_str() );

    while ( const COlo *fin = fiter.next() ) {
        POlo fout( fin->new_POlo() );
        fout->log_action( "mso "+expr );

        int nj = fin->nJ();
        vector<double> v(nj);
        for (int j=0; j<nj; j++)
            v[j] = T->tree_point_dbl( fiter.k(), j );

        vector<size_t> P = triv::sorted_indices( v );
        for (int j=0; j<nj; j++)
            fout->V.push_back( fin->copy_slice( P[j] ) );

        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Sort slices according to full z vector.

void NManip::slices_sort_by_z()
{
    FileIterator fiter(SFSel::instance()->sel());
    while ( const COlo *fin = fiter.next() ) {
        POlo fout( fin->new_POlo() );
        fout->log_action( "msoz" );
        for (int j=0; j<fin->nJ(); j++)
            fout->V.push_back( fin->copy_slice( j ) );
        sort( fout->V.begin(), fout->V.end(), []( const PSlice& E1, const PSlice& E2 ) {
                if( E1->z.size()!=E2->z.size() )
                    throw S("BUG: inconsistent z.size in CompareZ");
                int nz = E1->z.size();
                for( int iz=0; iz<nz; ++iz ){
                    if    ( E1->z[iz]->to_r() > E2->z[iz]->to_r() )
                        return false;
                    if    ( E1->z[iz]->to_r() < E2->z[iz]->to_r() )
                        return true;
                }
                return false;
            } );
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Change order of z coordinates.

void NManip::zcoords_reorder()
{ // EMBEDDED_DIALOG
    FileIterator fiter(SFSel::instance()->sel());
    int nzmin, nzmax = 0;
    char mod;
    string com;
    int num1=-1, num2=-1;  // nonsense initialization to suppress warning

    while ( const COlo *fin = fiter.next() ) {
        nzmax = std::max( (int)nzmax, (int)fin->ZCo.size() );
    }
    if       ( nzmax<=1 )
        throw S("nothing to exchange");
    else if ( nzmax==2 ) {
        mod = 'r';
        printf( "exchanging two z coordinates\n" );
        com = "exchanged two z coordinates";
        num1 = 1;
        nzmin = 2;
    } else {
        string sel;
        sel = wask( "Rotate(r,rr,...), revert(m), exchange(<number>)" );
        if      ( sel[0]=='r' ) {
            mod = 'r';
            num1 = 1;
            while ( sel.length() > num1 ) {
                if ( sel[num1++]!='r' )
                    throw S("invalid selection");
            }
            com = "rotated z coordinates by " + S(num1);
            nzmin = 2;
        } else if ( sel=="m" ) {
            mod = 'm';
            com = "reverted z coordinates";
            nzmin = 2;
        } else {
            if ( sscanf(sel.c_str(), "%i", &num1)!=1 )
                throw S("invalid selection");
            if ( num1>=nzmax )
                throw "no more than " + S(nzmax) + " z coordinates";
            num2 = iask( "Exchange with coordinate" );
            if ( num2>=nzmax )
                throw "no more than " + S(nzmax) + " z coordinates";
            else if ( num1==num2 )
                throw S("nothing to exchange");
            mod = 'x';
            com = "exchanged z coordinates "+S(num1)+" and "+S(num2);
            nzmin = std::max( num1, num2 )+1;
        }
    }

    fiter.reset();
    while ( const COlo *fin = fiter.next() ) {
        POlo fout( fin->new_POlo() );
        fout->log_action( com );
        for (int j=0; j<fin->nJ(); j++)
            fout->V.push_back( fin->copy_slice( j ) );

        int nz = fout->ZCo.size();
        if      ( nz<nzmin ) {
            cout << "WARNING nz<nzmin\n";
        } else if ( mod=='r' ) {
            for ( int irot=0; irot<num1; ++irot ) {
                for ( int iz=1; iz<nz; ++iz ) {
                    fout->ZCo[iz] = fin->ZCo[iz-1];
                }
                fout->ZCo[0] = fin->ZCo[nz-1];
            }
        } else if ( mod=='m' ) {
            for ( int iz=0; iz<nz; ++iz ) {
                fout->ZCo[iz] = fin->ZCo[nz-1-iz];
            }
        } else { // mod=='x'
            fout->ZCo[num1] = fin->ZCo[num2];
            fout->ZCo[num2] = fin->ZCo[num1];
        }

        for ( int j=0; j<fin->V.size(); j++ ) {
            if      ( nz<nzmin ) {
                // do nothing
            } else if ( mod=='r' ) {
                for ( int irot=0; irot<num1; ++irot ) {
                    for ( int iz=1; iz<nz; ++iz )
                        fout->V[j]->z[iz] = fin->V[j]->z[iz-1];
                    fout->V[j]->z[0] = fin->V[j]->z[nz-1];
                }
            } else if ( mod=='m' ) {
                for ( int iz=0; iz<nz; ++iz ) {
                    fout->V[j]->z[iz] = fin->V[j]->z[nz-1-iz];
                }
            } else  { // mod=='x'
                fout->V[j]->z[num1] = fin->V[j]->z[num2];
                fout->V[j]->z[num2] = fin->V[j]->z[num1];
            }
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Delete z coordinate.

void NManip::zcoord_delete()
{
    FileIterator fiter(SFSel::instance()->sel());
    string sel;
    sel = wask( "List of z coordinates to be removed" );
    if ( sel=="" )
        return;
    RNode Tsel = user_xaxparse( sel.c_str() );

    while ( const COlo *fin = fiter.next() ) {
        POlo fout( fin->new_POlo() );
        fout->log_action( "mz- "+sel );

        if ( !fin->nJ() )
            throw S("BUG: no spectra");
        for ( int j=0; j<fin->nJ(); j++ )
            fout->V.push_back( fin->copy_slice( j ) );

        RObjVecInt vsel = Tsel->to_index_list( fin->ZCo.size() );
        if (!vsel)
            throw "Invalid index list '"+sel+"'; max index is "+S(fin->ZCo.size()-1);
        if ( fin->nJ()>1 && fin->ZCo.size()-vsel->size()<1 )
            throw "Cannot delete all z coordinates since there is more than one slice";
        vector<int> lis = vsel->v;
        sort( lis.begin(), lis.end() );
        reverse( lis.begin(), lis.end() );

        int nRPar = fout->RPar.size();
        for ( int iv=0; iv<lis.size(); ++iv ) {
            int iz = lis[iv];
            fout->RPar.insert( fout->RPar.begin()+nRPar, fout->ZCo[iz] );
            fout->ZCo.erase( fout->ZCo.begin()+iz );
            double zval = 0;
            for ( int j=0; j<fin->nJ(); j++ ) {
                auto pz = PCAST<const CObjNum>(fin->z(j,iz));
                if( !pz )
                    throw "z"+S(iz)+" is not a number";
                zval += pz->to_r();
                fout->V[j]->z.erase( fout->V[j]->z.begin()+iz );
            }
            fout->RPar[nRPar].val = zval / fin->nJ();
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//! Break file into several files, according to z0.

void NManip::slices_break()
{
    FileIterator fiter(SFSel::instance()->sel());
    while ( const COlo *fin = fiter.next() ) {
        if ( !(fin->ZCo.size()) )
            throw S("no z coordinate");
        CCoord zco = fin->ZCo[0];

        // intermediate file, on which output files will be based:
        POlo ftmp( fin->new_POlo() );
        ftmp->V.clear();
        ftmp->log_action( "msb # eliminating " + zco.str_compact() );
        ftmp->ZCo.erase( ftmp->ZCo.begin() );
        int jout = 0;

        for ( int j=0; j<fin->nJ(); ) {
            POlo fout( ftmp->new_POlo() );
            fout->log_action( "msb" );
            double zval = fin->z(j,0)->to_r();
            fout->RPar.push_back( CParam( zco, zval ) );
            do {
                PSlice sout( fin->copy_slice(j) );
                sout->z.erase( sout->z.begin() );
                fout->V.push_back( move(sout) );
            } while ( ++j<fin->nJ() && fin->z(j,0)->to_r()==zval );
            fout->name += "_" + S(jout++);
            SMem::instance()->mem_store( move(fout) );
        }
        if ( SMem::instance()->overwrite )
            SMem::instance()->mem_del( fiter.k() );
    }
}


//**************************************************************************************************
//*  manipulations on files
//**************************************************************************************************

//! Merge files, concatenating spectra/curves.

void NManip::files_merge( const string& opts )
{
    if ( SFSel::instance()->sel().size()<=1 )
        throw S("won't merge less than 2 files");

    FileIterator fiter(SFSel::instance()->sel());
    // parse options:
    bool add_zk = false;
    if ( opts=="+" )
        add_zk = true;

    // Determine file type, which must be the same for all input files.
    const COlo *fin = fiter.next();
    POlo fout;
    if        ( dynamic_cast<const COld*>(fin) ) {
        fout = POld( new COld() );
        while ( fin = fiter.next() )
            if ( !dynamic_cast<const COld*>(fin) )
                throw S("cannot merge data and curve file");
    } else if ( dynamic_cast<const COlc*>(fin) ) {
        fout = POlc( new COlc() );
        while ( fin = fiter.next() )
            if ( !dynamic_cast<const COlc*>(fin) )
                throw S("cannot merge curve and data file");
    } else
        throw S("BUG: unexpected POlo type");

    // Check and set coordinates and other file-level attributes.
    fiter.reset();
    fin = fiter.next();
    fout->xco = fin->xco;
    fout->yco = fin->yco;
    fout->ZCo = fin->ZCo;
    if ( const COlc *fc = dynamic_cast<const COlc*>(fin) ) {
        COlc *fcout = dynamic_cast<COlc*>(fout.get());
        fcout->expr = fc->expr;
        fcout->evaMode = fc->evaMode;
        fcout->scrInpMode = fc->scrInpMode;
        fcout->nP = fc->nP;
        fcout->PCo = fc->PCo;
        fcout->T = fc->T;
    }
    while ( fin = fiter.next() ) {
        if ( fin->xco != fout->xco )
            throw S("different x coordinates");
        if ( fin->yco != fout->yco )
            throw S("different y coordinates");
        if ( fin->ZCo.size() != fout->ZCo.size() )
            throw S("different numer of z coordinates");
        for ( int i=0; i<fout->ZCo.size(); ++i )
            if ( fin->ZCo[i] != fout->ZCo[i] )
                throw "different z"+S(i)+" coordinate";
        if ( const COlc *fc = dynamic_cast<const COlc*>(fin) ) {
            const COlc *fcout = dynamic_cast<const COlc*>(fout.get());
            if ( fc->expr != fcout->expr  )
                throw S("different curve definitions");
            if ( fc->evaMode != fcout->evaMode )
                throw S("different curve evaluation mode");
            if ( fc->nP != fcout->nP )
                throw S("different number of curve parameters");
            if ( fc->PCo.size() != fcout->PCo.size() )
                throw S("inconsistent number of curve parameters");
        }
    }

    // check RPar, and prepare move to z if they differ:
    fiter.reset();
    vector<CParam> RPar( fiter.next()->RPar );
    int nR = RPar.size();
    vector<bool> RParDiff( nR, false );
    while ( (fin = fiter.next()) ) {
        if ( fin->RPar.size() != nR )
            throw S("different numer of r parameters");
        for ( int i=0; i<nR; ++i )
            if (fin->RPar[i].Co != RPar[i].Co )
                throw S("different r parameter");
        for ( int i=0; i<nR; ++i )
            if (fin->RPar[i].val != RPar[i].val )
                RParDiff[i] = true;
    }
    auto zinsert = fout->ZCo.begin();
    for ( int i=0; i<nR; ++i ) {
        if ( RParDiff[i] )
            fout->ZCo.insert( zinsert++, RPar[i].Co );
        else
            fout->RPar.push_back( RPar[i] );
    }
    if ( add_zk || !fout->ZCo.size() ) {
        // there must be at least one z coordinate
        fout->ZCo.insert( fout->ZCo.begin(), CCoord("number-of-input-file", "") );
        add_zk = true;
    }

    // file name:
    string fnam;
    fiter.reset();
    fnam = fiter.next()->name;
    while ( fin = fiter.next() ) {
        int n2=fin->name.size();
        if ( n2<fnam.size() )
            fnam.erase( fnam.begin()+n2, fnam.end() );
        for ( int i=0; i<fnam.size(); ++i )
            if (fnam[i]!=(fin->name)[i])
                fnam[i] = ' ';
    }
    for ( auto cp = fnam.begin(); cp != fnam.end(); ++cp )
        if (*cp==' ')
            fnam.erase(cp--);
    if (fnam == "") {
        static int callNo=0;
        fnam = str( format( "merge%03u" ) % callNo++ );
    }
    fout->name = fnam;

    // doc lines:
    fout->log_action( "mfj # " + fnam + " is merger of:" );
    fiter.reset();
    const COlo *f1 = fiter.next();
    fout->log_action("- " + f1->name);
    for ( int i=0; i<f1->lDoc.size(); ++i )
        fout->log_action("  " + f1->lDoc[i]);
    while ( (fin = fiter.next()) ) {
        fout->log_action("- " + fin->name);
        string line = "  ";
        for ( int i=0; i<fin->lDoc.size(); ++i ) {
            if ( i<f1->lDoc.size() && fin->lDoc[i]==f1->lDoc[i] ) {
                line += "=";
            } else {
                fout->log_action( line + fin->lDoc[i] );
                line = "  ";
            }
        }
    }

    // now merge the Zentries, and add new z's:
    fiter.reset();
    while ( fin = fiter.next() ) {
        for ( int j=0; j<fin->nJ(); j++ ) {
            PSlice sout( fin->copy_slice(j) );
            auto zinsert = sout->z.begin();
            for ( int i=0; i<nR; ++i )
                if ( RParDiff[i] )
                    sout->z.insert( zinsert++, PObjDbl( new CObjDbl( fin->RPar[i].val ) ) );
            if ( add_zk )
                sout->z.insert( sout->z.begin(), PObjInt( new CObjInt( fiter.iteration() ) ) );
            fout->V.push_back( move(sout) );
        }
    }

    if ( SMem::instance()->overwrite ) {
        SFSel::instance()->sel_delete();
        SMem::instance()->overwrite = false; // since we modified the file selection
    }
    SMem::instance()->mem_store( move(fout) );
}


//! Merge files with same number of spectra, concatenating spectra with same j.

void NManip::files_merge_pointwise()
{
    FileIterator fiter(SFSel::instance()->selD());
    // Copy first input file:
    const COld *fin = fiter.nextD();
    POld fout( fin->new_POld() );
    fout->copy_mainvec( fin );

    // Merge further input files:
    while ( fin = fiter.nextD() ) {
        if ( fin->nJ() != fout->nJ() )
            throw S("different # spectra");
        if ( fin->xco != fout->xco || fin->yco != fout->yco )
            throw S("different x/y coordinates");

        for ( int j=0; j<fout->nJ(); ++j ) {
            const CSpec *sin = fin->VS(j);
            CSpec* sout = fout->VS(j);
            if ( sin->z.size() != sout->z.size() )
                throw S("different # z entries");
            for ( int i=0; i<sout->z.size(); ++i )
                if ( sin->z[i] != sout->z[i] )
                    throw S("different # z values");
            if ( sin->has_dy() ^ sout->has_dy() )
                throw S("different on/off status of error bars");
            for ( int i=0; i<sin->size(); ++i ) {
                if ( sin->has_dy() )
                    sout->push_xyd( sin->x[i], sin->y[i], sin->dy[i] );
                else
                    sout->push_xy( sin->x[i], sin->y[i] );
            }
        }
        fout->log_action( "appended pointwise " + fin->name );
    }

    SMem::instance()->mem_store( move(fout) );
}


//**************************************************************************************************
//*  Interpolation
//**************************************************************************************************

//! Locate y in x' to determine y'(x'=y).

void NManip::interpolate()
{
    string expr = wask( "Interpolate in reference file" );
    if (expr=="")
        return;
    RNode T = user_xaxparse( expr.c_str() );

    FileIterator fiter(SFSel::instance()->selD());
    while ( const COld *fin = fiter.nextD() ) {
        int k = fiter.k();
        CContext ctx( k );
        int k2 = T->tree_val_idx( ctx, "K" );
        COld *f2 = SMem::instance()->mem_get_D( k2 );

        POld fout( fin->new_POld() );
        fout->copy_mainvec( fin );
        fout->log_action( "mi " + S(k2) );

        for ( int j=0; j<fin->nJ(); j++ ) {
            const CSpec *sin = fin ->VS(j);
            CSpec* s2  = f2  ->VS(j);
            CSpec* sout= fout->VS(j);
            int i2 = 0;
            if ( sin->y[0]<s2->x[0] )
                throw "File "+S(k)+", spec "+S(j)+": y[0] not in reference";
            for ( int i=0; i<sin->size(); i++ ) {
                double p = sin->y[i];
                if ( p==s2->x[i2] ) {
                    sout->y[i] = s2->y[i2];
                    continue;
                }
                while ( i2+1<s2->size() ) {
                    if ( p<=s2->x[i2+1] )
                        goto found_i2;
                    ++i2;
                }
                throw "File "+S(k)+", spec "+S(j)+": y["+S(i)+
                    "] not in reference";
            found_i2:
                double step = s2->x[i2+1] - s2->x[i2];
                if ( step <= 0 )
                    throw S("Reference not monotonous");
                sout->y[i] = ( s2->y[i2]*(s2->x[i2+1]-p) +
                               s2->y[i2+1]*(p-s2->x[i2]) ) / step;
            }
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}


//**************************************************************************************************
//*  Histogramming
//**************************************************************************************************

//! Create and fill histograms based on x values and their y weights.

void NManip::make_histogram()
{
    // Determine range:
    double xmin=0, xmax=0; // initialized to prevent warning
    int npts = 0;
    FileIterator fiter(SFSel::instance()->selD());
    while ( const COld *fin = fiter.nextD() ) {
        for ( int j=0; j<fin->nJ(); j++ ) {
            const CSpec *sin = fin ->VS(j);
            for ( int i=0; i<sin->size(); i++ ) {
                double x = sin->x[i];
                if ( !npts ) {
                    xmin = x;
                    xmax = x;
                } else if ( x<xmin ) {
                    xmin = x;
                } else if ( x>xmax ) {
                    xmax = x;
                }
                ++npts;
            }
        }
    }
    cout << "there are " << npts << " points extending from " << xmin <<
        " to " << xmax << "\n";

    // Obtain number of bins:
    static int nbin = 32;
    if ( npts<nbin/3 ) // adapt default
        npts = nbin/4;
    nbin = iask( "Number of bins", nbin );
    if ( nbin<2 )
        throw S("at least 2 bins required");

    // Set output grid:
    vector<double> xout(nbin);
    double dx = (xmax-xmin)/(nbin-1);
    for ( int i=0; i<nbin; ++i )
        xout[i] = xmin + dx*i;

    // Create and fill output histograms:
    fiter.reset();
    while ( const COld *fin = fiter.nextD() ) {
        POld fout( fin->new_POld() );
        fout->log_action("mhm " + S(nbin) );

        for ( int j=0; j<fin->nJ(); j++ ) {
            const CSpec *sin = fin->VS(j);
            PSpec sout( new CSpec() );
            sout->z = sin->z;
            sout->x = xout;
            sout->y = vector<double>(nbin, 0.);
            for ( int i=0; i<sin->size(); ++i ) {
                int iout = lrint( floor((sin->x[i]-xmin)/dx+0.5 ) );
                if ( iout<0 || iout>=nbin )
                    throw "BUG: invalid iout="+S(iout)+" for y="+S(sin->x[i]);
                sout->y[iout] += sin->y[i];
            }
            fout->V.push_back( move(sout) );
        }
        SMem::instance()->mem_store( move(fout), fiter.k() );
    }
}
